# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from nvshmem.bindings.device.numba import *

import cffi
from numba.core.extending import overload
from numba.types import int8, int16, int32, int64, uint8, uint16, uint32, uint64, float32, float64, float16, Array
from numba.cuda.types import bfloat16

__all__ = ["p", "g", "put", "get", "put_nbi", "get_nbi", "put_block", "get_block", "put_nbi_block", "get_nbi_block", "put_warp", "get_warp", "put_nbi_warp", "get_nbi_warp", "put_signal_block", "put_signal", "put_signal_nbi", "put_signal_warp", "put_signal_nbi_block", "put_signal_nbi_warp", "put_signal_nbi_warp"]

# TODO: create a global ffi object for other high level bindings to use
ffi = cffi.FFI()

def put(src: Array, dst: Array, pe: int32) -> None:
    """
    Copies data from local ``src`` to symmetric ``dst`` on PE ``pe``. 
    This is a thread-level operation.

    Args:
        - ``src`` (``Array``): Local source array on this PE to copy from.
        - ``dst`` (``Array``): Symmetric destination array on PE ``pe`` to copy to.
        - ``pe`` (``int``): PE to copy to.
    
    Note:
        When ``src`` and ``dst`` are of different sizes, only data of the smaller size is copied.
        ``src`` and ``dst`` must have the same data type.
    """
    pass

def get(dst: Array, src: Array, pe: int32) -> None:
    """
    Copies data from symmetric ``src`` on PE ``pe`` to local ``dst``.
    This is a thread-level operation.

    Args:
        - ``dst`` (``Array``): Local destination array on this PE to copy to.
        - ``src`` (``Array``): Symmetric source array from PE ``pe`` to copy from.
        - ``pe`` (``int``): PE to copy from.
    
    Note:
        When ``src`` and ``dst`` are of different sizes, only data of the smaller size is copied.
        ``src`` and ``dst`` must have the same data type.
    """
    pass

def put_nbi(src: Array, dst: Array, pe: int32) -> None:
    """
    Non-blockingly copies data from local ``src`` to symmetric ``dst`` on PE ``pe``.
    This is a thread-level operation.

    Args:
        - ``src`` (``Array``): Local source array on this PE to copy from.
        - ``dst`` (``Array``): Symmetric destination array on PE ``pe`` to copy to.
        - ``pe`` (``int``): PE to copy to.
    
    Note:
        When ``src`` and ``dst`` are of different sizes, only data of the smaller size is copied.
        ``src`` and ``dst`` must have the same data type.
    """
    pass

def get_nbi(dst: Array, src: Array, pe: int32) -> None:
    """
    Non-blockingly copies data from symmetric ``src`` on PE ``pe`` to local ``dst``.
    This is a thread-level operation.

    Args:
        - ``dst`` (``Array``): Local destination array on this PE to copy to.
        - ``src`` (``Array``): Symmetric source array from PE ``pe`` to copy from.
        - ``pe`` (``int``): PE to copy from.
    
    Note:
        When ``src`` and ``dst`` are of different sizes, only data of the smaller size is copied.
        ``src`` and ``dst`` must have the same data type.
    """
    pass

def put_block(src: Array, dst: Array, pe: int32) -> None:
    """
    Copies from local ``src`` to symmetric ``dst`` on PE ``pe``.
    This is a CTA-level operation. All threads in the CTA must call this function with the same arguments.

    Args:
        - ``src`` (``Array``): Local source array on this PE to copy from.
        - ``dst`` (``Array``): Symmetric destination array on PE ``pe`` to copy to.
        - ``pe`` (``int``): PE to copy to.
    
    Note:
        When ``src`` and ``dst`` are of different sizes, only data of the smaller size is copied.
        ``src`` and ``dst`` must have the same data type.
    """
    pass

def get_block(dst: Array, src: Array, pe: int32) -> None:
    """
    Copies from symmetric ``src`` on PE ``pe`` to local ``dst``.
    This is a CTA-level operation. All threads in the CTA must call this function with the same arguments.

    Args:
        - ``dst`` (``Array``): Local destination array on this PE to copy to.
        - ``src`` (``Array``): Symmetric source array from PE ``pe`` to copy from.
        - ``pe`` (``int``): PE to copy from.
    
    Note:
        When ``src`` and ``dst`` are of different sizes, only data of the smaller size is copied.
        ``src`` and ``dst`` must have the same data type.
    """
    pass

def put_nbi_block(src: Array, dst: Array, pe: int32) -> None:
    """
    Non-blockingly copies from local ``src`` to symmetric ``dst`` on PE ``pe``.
    This is a CTA-level operation. All threads in the CTA must call this function with the same arguments.

    Args:
        - ``src`` (``Array``): Local source array on this PE to copy from.
        - ``dst`` (``Array``): Symmetric destination array on PE ``pe`` to copy to.
        - ``pe`` (``int``): PE to copy to.
    """
    pass

def get_nbi_block(dst: Array, src: Array, pe: int32) -> None:
    """
    Non-blockingly copies from symmetric ``src`` on PE ``pe`` to local ``dst``.
    This is a CTA-level operation. All threads in the CTA must call this function with the same arguments.

    Args:
        - ``dst`` (``Array``): Local destination array on this PE to copy to.
        - ``src`` (``Array``): Symmetric source array from PE ``pe`` to copy from.
        - ``pe`` (``int``): PE to copy from.
    """
    pass

def put_warp(src: Array, dst: Array, pe: int32) -> None:
    """
    Copies from local ``src`` to symmetric ``dst`` on PE ``pe``.
    This is a warp-level operation. All threads in the warp must call this function with the same arguments.

    Args:
        - ``src`` (``Array``): Local source array on this PE to copy from.
        - ``dst`` (``Array``): Symmetric destination array on PE ``pe`` to copy to.
        - ``pe`` (``int``): PE to copy to.
    
    Note:
        When ``src`` and ``dst`` are of different sizes, only data of the smaller size is copied.
        ``src`` and ``dst`` must have the same data type.
    """
    pass

def get_warp(dst: Array, src: Array, pe: int32) -> None:
    """
    Copies from symmetric ``src`` on PE ``pe`` to local ``dst``.
    This is a warp-level operation. All threads in the warp must call this function with the same arguments.

    Args:
        - ``dst`` (``Array``): Local destination array on this PE to copy to.
        - ``src`` (``Array``): Symmetric source array from PE ``pe`` to copy from.
        - ``pe`` (``int``): PE to copy from.
    
    Note:
        When ``src`` and ``dst`` are of different sizes, only data of the smaller size is copied.
        ``src`` and ``dst`` must have the same data type.
    """
    pass

def put_nbi_warp(src: Array, dst: Array, pe: int32) -> None:
    """
    Non-blockingly copies from local ``src`` to symmetric ``dst`` on PE ``pe``.
    This is a warp-level operation. All threads in the warp must call this function with the same arguments.

    Args:
        - ``src`` (``Array``): Local source array on this PE to copy from.
        - ``dst`` (``Array``): Symmetric destination array on PE ``pe`` to copy to.
        - ``pe`` (``int``): PE to copy to.
    
    Note:
        When ``src`` and ``dst`` are of different sizes, only data of the smaller size is copied.
        ``src`` and ``dst`` must have the same data type.
    """
    pass

def get_nbi_warp(dst: Array, src: Array, pe: int32) -> None:
    """
    Non-blockingly copies from symmetric ``src`` on PE ``pe`` to local ``dst``.
    This is a warp-level operation. All threads in the warp must call this function with the same arguments.

    Args:
        - ``dst`` (``Array``): Local destination array on this PE to copy to.
        - ``src`` (``Array``): Symmetric source array from PE ``pe`` to copy from.
        - ``pe`` (``int``): PE to copy from.
    
    Note:
        When ``src`` and ``dst`` are of different sizes, only data of the smaller size is copied.
        ``src`` and ``dst`` must have the same data type.
    """
    pass

# put variations
@overload(put_block)
def put_block_ol(dst, src, pe):
    if dst == Array(dtype=int8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int8_put_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int16_put_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int32_put_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int64_put_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint8_put_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint16_put_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint32_put_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint64_put_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            float_put_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            double_put_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            half_put_block(dstptr, srcptr, nelems, pe)
        return impl

@overload(put_nbi_block)
def put_nbi_block_ol(dst, src, pe):
    if dst == Array(dtype=int8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int8_put_nbi_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int16_put_nbi_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int32_put_nbi_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int64_put_nbi_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint8_put_nbi_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint16_put_nbi_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint32_put_nbi_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint64_put_nbi_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            float_put_nbi_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            double_put_nbi_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            half_put_nbi_block(dstptr, srcptr, nelems, pe)
        return impl


@overload(put_warp)
def put_warp_ol(dst, src, pe):
    if dst == Array(dtype=int8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int8_put_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int16_put_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int32_put_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int64_put_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint8_put_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint16_put_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint32_put_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint64_put_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            float_put_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            double_put_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            half_put_warp(dstptr, srcptr, nelems, pe)
        return impl

@overload(put_nbi_warp)
def put_nbi_warp_ol(dst, src, pe):
    if dst == Array(dtype=int8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int8_put_nbi_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int16_put_nbi_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int32_put_nbi_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int64_put_nbi_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint8_put_nbi_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint16_put_nbi_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint32_put_nbi_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint64_put_nbi_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            float_put_nbi_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            double_put_nbi_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            half_put_nbi_warp(dstptr, srcptr, nelems, pe)
        return impl


@overload(put)
def put_ol(dst, src, pe):
    if dst == Array(dtype=int8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int8_put(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int16_put(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int32_put(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int64_put(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint8_put(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint16_put(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint32_put(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint64_put(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            float_put(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            double_put(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            half_put(dstptr, srcptr, nelems, pe)
        return impl

@overload(put_nbi)
def put_nbi_ol(dst, src, pe):
    if dst == Array(dtype=int8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int8_put_nbi(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int16_put_nbi(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int32_put_nbi(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int64_put_nbi(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint8_put_nbi(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint16_put_nbi(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint32_put_nbi(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint64_put_nbi(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            float_put_nbi(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            double_put_nbi(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            half_put_nbi(dstptr, srcptr, nelems, pe)
        return impl



# get variations
@overload(get_block)
def get_block_ol(dst, src, pe):
    if dst == Array(dtype=int8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int8_get_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int16_get_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int32_get_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int64_get_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint8_get_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint16_get_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint32_get_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint64_get_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            float_get_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            double_get_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            half_get_block(dstptr, srcptr, nelems, pe)
        return impl

@overload(get_nbi_block)
def get_nbi_block_ol(dst, src, pe):
    if dst == Array(dtype=int8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int8_get_nbi_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int16_get_nbi_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int32_get_nbi_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int64_get_nbi_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint8_get_nbi_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint16_get_nbi_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint32_get_nbi_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint64_get_nbi_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            float_get_nbi_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            double_get_nbi_block(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            half_get_nbi_block(dstptr, srcptr, nelems, pe)
        return impl


@overload(get_warp)
def get_warp_ol(dst, src, pe):
    if dst == Array(dtype=int8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int8_get_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int16_get_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int32_get_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int64_get_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint8_get_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint16_get_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint32_get_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint64_get_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            float_get_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            double_get_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            half_get_warp(dstptr, srcptr, nelems, pe)
        return impl

@overload(get_nbi_warp)
def get_nbi_warp_ol(dst, src, pe):
    if dst == Array(dtype=int8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int8_get_nbi_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int16_get_nbi_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int32_get_nbi_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int64_get_nbi_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint8_get_nbi_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint16_get_nbi_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint32_get_nbi_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint64_get_nbi_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            float_get_nbi_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            double_get_nbi_warp(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            half_get_nbi_warp(dstptr, srcptr, nelems, pe)
        return impl


@overload(get)
def get_ol(dst, src, pe):
    if dst == Array(dtype=int8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int8_get(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int16_get(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int32_get(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int64_get(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint8_get(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint16_get(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint32_get(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint64_get(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            float_get(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            double_get(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            half_get(dstptr, srcptr, nelems, pe)
        return impl

@overload(get_nbi)
def get_nbi_ol(dst, src, pe):
    if dst == Array(dtype=int8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int8_get_nbi(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int16_get_nbi(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int32_get_nbi(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=int64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            int64_get_nbi(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint8_get_nbi(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint16_get_nbi(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint32_get_nbi(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=uint64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            uint64_get_nbi(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            float_get_nbi(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            double_get_nbi(dstptr, srcptr, nelems, pe)
        return impl
    elif dst == Array(dtype=float16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            nelems = min(dst.size, src.size)
            half_get_nbi(dstptr, srcptr, nelems, pe)
        return impl



# put_signal variations
def put_signal_block():
    """
    Put data with a signal operation from ``src`` to ``dst`` on PE ``pe`` in a  manner at a  scope. Device initiated.

    Signal variables must be an ``Array`` of dtype ``int64`` (8 bytes) allocated by or registered with NVSHMEM4Py.
    Supported signal operations are ``SignalOp.SIGNAL_SET`` and ``SignalOp.SIGNAL_ADD``.

    Args:
        - ``dst`` (``Array``): Destination symmetric array on remote PE.
        - ``src`` (``Array``): Source symmetric array on this PE.
        - ``signal_var`` (``Array``): Symmetric signal variable (dtype ``int64``).
        - ``signal_val`` (``int``): Signal value.
        - ``signal_op`` (``SignalOp``): Signal operation type.
        - ``pe`` (``int``): Target PE to put to.
    """
    pass
@overload(put_signal_block)
def put_signal_block_ol(dst, src, signal_var, signal_val, signal_op, pe):
    if dst == Array(dtype=int8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            int8_put_signal_block(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=int16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            int16_put_signal_block(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=int32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            int32_put_signal_block(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=int64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            int64_put_signal_block(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=uint8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            uint8_put_signal_block(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=uint16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            uint16_put_signal_block(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=uint32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            uint32_put_signal_block(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=uint64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            uint64_put_signal_block(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=float32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            float_put_signal_block(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=float64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            double_put_signal_block(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=float16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            half_put_signal_block(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl

def put_signal_nbi_block():
    """
    Put data with a signal operation from ``src`` to ``dst`` on PE ``pe`` in a _nbi manner at a  scope. Device initiated.

    Signal variables must be an ``Array`` of dtype ``int64`` (8 bytes) allocated by or registered with NVSHMEM4Py.
    Supported signal operations are ``SignalOp.SIGNAL_SET`` and ``SignalOp.SIGNAL_ADD``.

    Args:
        - ``dst`` (``Array``): Destination symmetric array on remote PE.
        - ``src`` (``Array``): Source symmetric array on this PE.
        - ``signal_var`` (``Array``): Symmetric signal variable (dtype ``int64``).
        - ``signal_val`` (``int``): Signal value.
        - ``signal_op`` (``SignalOp``): Signal operation type.
        - ``pe`` (``int``): Target PE to put to.
    """
    pass
@overload(put_signal_nbi_block)
def put_signal_nbi_block_ol(dst, src, signal_var, signal_val, signal_op, pe):
    if dst == Array(dtype=int8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            int8_put_signal_nbi_block(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=int16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            int16_put_signal_nbi_block(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=int32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            int32_put_signal_nbi_block(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=int64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            int64_put_signal_nbi_block(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=uint8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            uint8_put_signal_nbi_block(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=uint16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            uint16_put_signal_nbi_block(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=uint32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            uint32_put_signal_nbi_block(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=uint64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            uint64_put_signal_nbi_block(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=float32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            float_put_signal_nbi_block(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=float64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            double_put_signal_nbi_block(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=float16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            half_put_signal_nbi_block(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl


def put_signal_warp():
    """
    Put data with a signal operation from ``src`` to ``dst`` on PE ``pe`` in a  manner at a  scope. Device initiated.

    Signal variables must be an ``Array`` of dtype ``int64`` (8 bytes) allocated by or registered with NVSHMEM4Py.
    Supported signal operations are ``SignalOp.SIGNAL_SET`` and ``SignalOp.SIGNAL_ADD``.

    Args:
        - ``dst`` (``Array``): Destination symmetric array on remote PE.
        - ``src`` (``Array``): Source symmetric array on this PE.
        - ``signal_var`` (``Array``): Symmetric signal variable (dtype ``int64``).
        - ``signal_val`` (``int``): Signal value.
        - ``signal_op`` (``SignalOp``): Signal operation type.
        - ``pe`` (``int``): Target PE to put to.
    """
    pass
@overload(put_signal_warp)
def put_signal_warp_ol(dst, src, signal_var, signal_val, signal_op, pe):
    if dst == Array(dtype=int8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            int8_put_signal_warp(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=int16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            int16_put_signal_warp(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=int32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            int32_put_signal_warp(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=int64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            int64_put_signal_warp(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=uint8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            uint8_put_signal_warp(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=uint16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            uint16_put_signal_warp(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=uint32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            uint32_put_signal_warp(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=uint64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            uint64_put_signal_warp(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=float32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            float_put_signal_warp(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=float64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            double_put_signal_warp(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=float16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            half_put_signal_warp(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl

def put_signal_nbi_warp():
    """
    Put data with a signal operation from ``src`` to ``dst`` on PE ``pe`` in a _nbi manner at a  scope. Device initiated.

    Signal variables must be an ``Array`` of dtype ``int64`` (8 bytes) allocated by or registered with NVSHMEM4Py.
    Supported signal operations are ``SignalOp.SIGNAL_SET`` and ``SignalOp.SIGNAL_ADD``.

    Args:
        - ``dst`` (``Array``): Destination symmetric array on remote PE.
        - ``src`` (``Array``): Source symmetric array on this PE.
        - ``signal_var`` (``Array``): Symmetric signal variable (dtype ``int64``).
        - ``signal_val`` (``int``): Signal value.
        - ``signal_op`` (``SignalOp``): Signal operation type.
        - ``pe`` (``int``): Target PE to put to.
    """
    pass
@overload(put_signal_nbi_warp)
def put_signal_nbi_warp_ol(dst, src, signal_var, signal_val, signal_op, pe):
    if dst == Array(dtype=int8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            int8_put_signal_nbi_warp(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=int16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            int16_put_signal_nbi_warp(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=int32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            int32_put_signal_nbi_warp(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=int64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            int64_put_signal_nbi_warp(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=uint8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            uint8_put_signal_nbi_warp(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=uint16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            uint16_put_signal_nbi_warp(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=uint32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            uint32_put_signal_nbi_warp(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=uint64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            uint64_put_signal_nbi_warp(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=float32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            float_put_signal_nbi_warp(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=float64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            double_put_signal_nbi_warp(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=float16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            half_put_signal_nbi_warp(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl


def put_signal():
    """
    Put data with a signal operation from ``src`` to ``dst`` on PE ``pe`` in a  manner at a  scope. Device initiated.

    Signal variables must be an ``Array`` of dtype ``int64`` (8 bytes) allocated by or registered with NVSHMEM4Py.
    Supported signal operations are ``SignalOp.SIGNAL_SET`` and ``SignalOp.SIGNAL_ADD``.

    Args:
        - ``dst`` (``Array``): Destination symmetric array on remote PE.
        - ``src`` (``Array``): Source symmetric array on this PE.
        - ``signal_var`` (``Array``): Symmetric signal variable (dtype ``int64``).
        - ``signal_val`` (``int``): Signal value.
        - ``signal_op`` (``SignalOp``): Signal operation type.
        - ``pe`` (``int``): Target PE to put to.
    """
    pass
@overload(put_signal)
def put_signal_ol(dst, src, signal_var, signal_val, signal_op, pe):
    if dst == Array(dtype=int8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            int8_put_signal(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=int16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            int16_put_signal(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=int32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            int32_put_signal(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=int64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            int64_put_signal(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=uint8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            uint8_put_signal(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=uint16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            uint16_put_signal(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=uint32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            uint32_put_signal(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=uint64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            uint64_put_signal(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=float32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            float_put_signal(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=float64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            double_put_signal(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=float16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            half_put_signal(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl

def put_signal_nbi():
    """
    Put data with a signal operation from ``src`` to ``dst`` on PE ``pe`` in a _nbi manner at a  scope. Device initiated.

    Signal variables must be an ``Array`` of dtype ``int64`` (8 bytes) allocated by or registered with NVSHMEM4Py.
    Supported signal operations are ``SignalOp.SIGNAL_SET`` and ``SignalOp.SIGNAL_ADD``.

    Args:
        - ``dst`` (``Array``): Destination symmetric array on remote PE.
        - ``src`` (``Array``): Source symmetric array on this PE.
        - ``signal_var`` (``Array``): Symmetric signal variable (dtype ``int64``).
        - ``signal_val`` (``int``): Signal value.
        - ``signal_op`` (``SignalOp``): Signal operation type.
        - ``pe`` (``int``): Target PE to put to.
    """
    pass
@overload(put_signal_nbi)
def put_signal_nbi_ol(dst, src, signal_var, signal_val, signal_op, pe):
    if dst == Array(dtype=int8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            int8_put_signal_nbi(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=int16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            int16_put_signal_nbi(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=int32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            int32_put_signal_nbi(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=int64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=int64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            int64_put_signal_nbi(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=uint8, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint8, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            uint8_put_signal_nbi(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=uint16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            uint16_put_signal_nbi(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=uint32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            uint32_put_signal_nbi(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=uint64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=uint64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            uint64_put_signal_nbi(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=float32, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float32, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            float_put_signal_nbi(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=float64, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float64, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            double_put_signal_nbi(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl
    elif dst == Array(dtype=float16, ndim=dst.ndim, layout=dst.layout) and src == Array(dtype=float16, ndim=src.ndim, layout=src.layout):
        def impl(dst, src, signal_var, signal_val, signal_op, pe):
            dstptr = ffi.from_buffer(dst)
            srcptr = ffi.from_buffer(src)
            signal_varptr = ffi.from_buffer(signal_var)
            nelems = min(dst.size, src.size)
            half_put_signal_nbi(dstptr, srcptr, nelems, signal_varptr, signal_val, signal_op, pe)
        return impl



# p variations
def p():
    """
    Put immediate data from src to dst on PE pe. Device initiated.

    ``src`` must be a scalar value, passed as a symmetric Array of size 1.

    Args:
        - ``dst`` (``Array``): Destination symmetric array on remote PE.
        - ``src`` (``Array``): Scalar source value (size-1 array) on this PE.
        - ``pe`` (``int``): PE to put to.
    """
    pass
@overload(p)
def p_ol(dst, src, pe):
    if dst == Array(dtype=int8, ndim=dst.ndim, layout=dst.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            int8_p(dstptr, src, pe)
        return impl
    elif dst == Array(dtype=int16, ndim=dst.ndim, layout=dst.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            int16_p(dstptr, src, pe)
        return impl
    elif dst == Array(dtype=int32, ndim=dst.ndim, layout=dst.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            int32_p(dstptr, src, pe)
        return impl
    elif dst == Array(dtype=int64, ndim=dst.ndim, layout=dst.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            int64_p(dstptr, src, pe)
        return impl
    elif dst == Array(dtype=uint8, ndim=dst.ndim, layout=dst.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            uint8_p(dstptr, src, pe)
        return impl
    elif dst == Array(dtype=uint16, ndim=dst.ndim, layout=dst.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            uint16_p(dstptr, src, pe)
        return impl
    elif dst == Array(dtype=uint32, ndim=dst.ndim, layout=dst.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            uint32_p(dstptr, src, pe)
        return impl
    elif dst == Array(dtype=uint64, ndim=dst.ndim, layout=dst.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            uint64_p(dstptr, src, pe)
        return impl
    elif dst == Array(dtype=float32, ndim=dst.ndim, layout=dst.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            float_p(dstptr, src, pe)
        return impl
    elif dst == Array(dtype=float64, ndim=dst.ndim, layout=dst.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            double_p(dstptr, src, pe)
        return impl
    elif dst == Array(dtype=float16, ndim=dst.ndim, layout=dst.layout):
        def impl(dst, src, pe):
            dstptr = ffi.from_buffer(dst)
            half_p(dstptr, src, pe)
        return impl

def g():
    """
    Get immediate data from src on PE pe to local dst. Device initiated.

    ``src`` must be a scalar value, passed as a symmetric Array of size 1.

    Args:
        - ``src`` (``Array``): Source symmetric array (size-1) on remote PE.
        - ``pe`` (``int``): PE to get from.
    Returns:
        Scalar value retrieved from ``src`` on PE ``pe``.
    """
    pass
@overload(g)
def g_ol(src, pe):
    if src == Array(dtype=int8, ndim=src.ndim, layout=src.layout):
        def impl(src, pe):
            srcptr = ffi.from_buffer(src)
            return int8_g(srcptr, pe)
        return impl
    elif src == Array(dtype=int16, ndim=src.ndim, layout=src.layout):
        def impl(src, pe):
            srcptr = ffi.from_buffer(src)
            return int16_g(srcptr, pe)
        return impl
    elif src == Array(dtype=int32, ndim=src.ndim, layout=src.layout):
        def impl(src, pe):
            srcptr = ffi.from_buffer(src)
            return int32_g(srcptr, pe)
        return impl
    elif src == Array(dtype=int64, ndim=src.ndim, layout=src.layout):
        def impl(src, pe):
            srcptr = ffi.from_buffer(src)
            return int64_g(srcptr, pe)
        return impl
    elif src == Array(dtype=uint8, ndim=src.ndim, layout=src.layout):
        def impl(src, pe):
            srcptr = ffi.from_buffer(src)
            return uint8_g(srcptr, pe)
        return impl
    elif src == Array(dtype=uint16, ndim=src.ndim, layout=src.layout):
        def impl(src, pe):
            srcptr = ffi.from_buffer(src)
            return uint16_g(srcptr, pe)
        return impl
    elif src == Array(dtype=uint32, ndim=src.ndim, layout=src.layout):
        def impl(src, pe):
            srcptr = ffi.from_buffer(src)
            return uint32_g(srcptr, pe)
        return impl
    elif src == Array(dtype=uint64, ndim=src.ndim, layout=src.layout):
        def impl(src, pe):
            srcptr = ffi.from_buffer(src)
            return uint64_g(srcptr, pe)
        return impl
    elif src == Array(dtype=float32, ndim=src.ndim, layout=src.layout):
        def impl(src, pe):
            srcptr = ffi.from_buffer(src)
            return float_g(srcptr, pe)
        return impl
    elif src == Array(dtype=float64, ndim=src.ndim, layout=src.layout):
        def impl(src, pe):
            srcptr = ffi.from_buffer(src)
            return double_g(srcptr, pe)
        return impl
    elif src == Array(dtype=float16, ndim=src.ndim, layout=src.layout):
        def impl(src, pe):
            srcptr = ffi.from_buffer(src)
            return half_g(srcptr, pe)
        return impl


# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
#
# This code was automatically generated from NVSHMEM with version 3.3.0. 
# Modify it directly at your own risk.


from libc.stdint cimport intptr_t

from .cynvshmem cimport *


###############################################################################
# Types
###############################################################################

ctypedef nvshmemx_uniqueid_args_v1 uniqueid_args
ctypedef nvshmem_team_config_v2 team_config
ctypedef nvshmemx_init_args_v2 init_args
ctypedef nvshmemx_init_attr_v2 init_attr

ctypedef cudaStream_t Stream
ctypedef CUmodule Module
ctypedef CUlibrary Library



###############################################################################
# Enum
###############################################################################

ctypedef nvshmemx_signal_op_t _Signal_op
ctypedef nvshmemx_tma_policy_t _Tma_policy
ctypedef nvshmemx_smem_amount_t _Smem_amount
ctypedef nvshmemx_cmp_type_t _Cmp_type
ctypedef nvshmemx_thread_support_t _Thread_support
ctypedef nvshmemx_proxy_status_t _Proxy_status
ctypedef nvshmemx_init_status_t _Init_status
ctypedef nvshmemx_qp_handle_index_t _Qp_handle_index
ctypedef nvshmem_pe_index_t _Pe_index
ctypedef nvshmem_team_id_t _Team_id
ctypedef nvshmemx_status _Status
ctypedef flags _Flags


###############################################################################
# Functions
###############################################################################

cpdef barrier(int32_t team)
cpdef void barrier_all() except*
cpdef int init_status() except? 0
cpdef int my_pe() except? -1
cpdef int n_pes() except? -1
cpdef void info_get_version(intptr_t major, intptr_t minor) except*
cpdef void vendor_get_version_info(intptr_t major, intptr_t minor, intptr_t patch) except*
cpdef intptr_t malloc(size_t size) except? 0
cpdef intptr_t calloc(size_t count, size_t size) except? 0
cpdef intptr_t align(size_t count, size_t size) except? 0
cpdef void free(intptr_t ptr) except*
cpdef intptr_t ptr(intptr_t dest, int pe) except? 0
cpdef intptr_t mc_ptr(int32_t team, intptr_t ptr) except? 0
cpdef int team_my_pe(int32_t team) except? -1
cpdef int team_n_pes(int32_t team) except? -1
cpdef void team_get_config(int32_t team, intptr_t config) except*
cpdef int team_translate_pe(int32_t src_team, int src_pe, int32_t dest_team) except? -1
cpdef team_split_strided(int32_t parent_team, int pe_start, int pe_stride, int pe_size, intptr_t config, long config_mask, intptr_t new_team)
cpdef team_get_uniqueid(intptr_t uniqueid)
cpdef team_init(intptr_t team, intptr_t config, long config_mask, int npes, int pe_idx_in_team)
cpdef team_split_2d(int32_t parent_team, int xrange, intptr_t xaxis_config, long xaxis_mask, intptr_t xaxis_team, intptr_t yaxis_config, long yaxis_mask, intptr_t yaxis_team)
cpdef void team_destroy(int32_t team) except*
cpdef bfloat16_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef half_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef float_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef double_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef char_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef short_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef schar_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef int_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef long_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef longlong_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef int8_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef int16_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef int32_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef int64_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef uint8_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef uint16_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef uint32_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef uint64_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef size_alltoall_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef barrier_on_stream(int32_t team, intptr_t stream)
cpdef void barrier_all_on_stream(intptr_t stream) except*
cpdef int team_sync_on_stream(int32_t team, intptr_t stream) except? 0
cpdef void sync_all_on_stream(intptr_t stream) except*
cpdef bfloat16_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)
cpdef half_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)
cpdef float_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)
cpdef double_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)
cpdef char_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)
cpdef short_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)
cpdef schar_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)
cpdef int_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)
cpdef long_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)
cpdef longlong_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)
cpdef int8_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)
cpdef int16_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)
cpdef int32_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)
cpdef int64_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)
cpdef uint8_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)
cpdef uint16_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)
cpdef uint32_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)
cpdef uint64_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)
cpdef size_broadcast_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, int pe_root, intptr_t stream)
cpdef bfloat16_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef half_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef float_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef double_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef char_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef short_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef schar_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef int_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef long_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef longlong_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef int8_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef int16_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef int32_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef int64_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef uint8_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef uint16_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef uint32_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef uint64_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef size_fcollect_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nelem, intptr_t stream)
cpdef int8_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int16_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int32_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int64_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef uint8_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef uint16_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef uint32_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef uint64_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef size_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef char_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef schar_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef short_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef long_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef longlong_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef bfloat16_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef half_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef float_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef double_max_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int8_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int16_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int32_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int64_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef uint8_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef uint16_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef uint32_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef uint64_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef size_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef char_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef schar_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef short_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef long_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef longlong_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef bfloat16_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef half_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef float_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef double_min_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int8_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int16_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int32_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int64_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef uint8_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef uint16_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef uint32_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef uint64_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef size_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef char_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef schar_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef short_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef long_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef longlong_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef bfloat16_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef half_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef float_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef double_sum_reduce_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int8_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int16_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int32_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int64_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef uint8_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef uint16_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef uint32_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef uint64_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef size_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef char_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef schar_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef short_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef long_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef longlong_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef bfloat16_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef half_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef float_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef double_max_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int8_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int16_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int32_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int64_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef uint8_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef uint16_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef uint32_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef uint64_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef size_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef char_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef schar_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef short_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef long_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef longlong_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef bfloat16_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef half_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef float_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef double_min_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int8_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int16_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int32_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int64_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef uint8_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef uint16_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef uint32_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef uint64_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef size_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef char_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef schar_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef short_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef int_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef long_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef longlong_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef bfloat16_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef half_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef float_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef double_sum_reducescatter_on_stream(int32_t team, intptr_t dest, intptr_t src, size_t nreduce, intptr_t stream)
cpdef hostlib_init_attr(unsigned int flags, intptr_t attr)
cpdef void hostlib_finalize() except*
cpdef set_attr_uniqueid_args(int myrank, int nranks, intptr_t uniqueid, intptr_t attr)
cpdef set_attr_mpi_comm_args(intptr_t mpi_comm, intptr_t nvshmem_attr)
cpdef get_uniqueid(intptr_t uniqueid)
cpdef int cumodule_init(intptr_t module) except? -1
cpdef int cumodule_finalize(intptr_t module) except? -1
cpdef intptr_t buffer_register_symmetric(intptr_t buf_ptr, size_t size, int flags) except? 0
cpdef int buffer_unregister_symmetric(intptr_t mmap_ptr, size_t size) except? 0
cpdef int culibrary_init(intptr_t library) except? -1
cpdef int culibrary_finalize(intptr_t library) except? -1
cpdef void putmem_on_stream(intptr_t dest, intptr_t source, size_t bytes, int pe, intptr_t cstrm) except*
cpdef void putmem_signal_on_stream(intptr_t dest, intptr_t source, size_t bytes, intptr_t sig_addr, uint64_t signal, int sig_op, int pe, intptr_t cstrm) except*
cpdef void getmem_on_stream(intptr_t dest, intptr_t source, size_t bytes, int pe, intptr_t cstrm) except*
cpdef void quiet_on_stream(intptr_t cstrm) except*
cpdef void signal_op_on_stream(intptr_t sig_addr, uint64_t signal, int sig_op, int pe, intptr_t cstrm) except*
cpdef void signal_wait_until_on_stream(intptr_t sig_addr, int cmp, uint64_t cmp_value, intptr_t cstream) except*


# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0

from cuda.pathfinder import find_nvidia_header_directory

from numba import cuda

import os
import warnings

from nvshmem.core.nvshmem_types import NvshmemWarning

if os.path.exists(os.path.join(os.path.dirname(__file__), "_numbast.py")):
    from ._numbast import *

    INCLUDE_PATH = find_nvidia_header_directory("nvshmem")
    if "nvshmem.h" not in os.listdir(INCLUDE_PATH):
        raise RuntimeError("nvshmem.h not found, package may not be properly installed")

    if not os.path.exists(INCLUDE_PATH):
        raise RuntimeError(
            f"NVSHMEM headers not found at {INCLUDE_PATH}. Please confirm that nvshmem is installed correctly.")

    CCCL_INCLUDE_PATH = find_nvidia_header_directory("cccl")

    if not os.path.exists(CCCL_INCLUDE_PATH):
        raise RuntimeError(
            f"CCCL headers not found at {CCCL_INCLUDE_PATH}. Please confirm that cccl is installed correctly.")

    # Path to this folder to look for entry point file
    this_folder = os.path.dirname(os.path.abspath(__file__))
    if not os.path.exists(os.path.join(this_folder, "entry_point.h")):
        raise RuntimeError("entry_point.h not found, package may not be properly installed")

    cuda.config.CUDA_NVRTC_EXTRA_SEARCH_PATHS = ":".join([INCLUDE_PATH, CCCL_INCLUDE_PATH, this_folder])

else:
    warnings.warn("Numba device bindings are not enabled", NvshmemWarning)
    _numbast = None



# Copyright (c) 2025-2026, NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
"""
__init__.py for nvshmem Python package
"""
import os

__package_name__ = os.getenv("PACKAGE_NAME", "nvshmem4py")

# Version is autogenerated by setuptools_scm
from .version import version, __version__, __version_tuple__
